import dearpygui.dearpygui as dpg
from .text import add_header
from ..constants import TEXT_COLOR, EMPHASIS_COLOR, SECONDARY_COLOR, WHITE


class TaskManagerWindow:
    
    
    def __init__(self):
        
        self.window_tag = dpg.generate_uuid()
    
        with dpg.window(
            label="Task Queue", 
            width=400, 
            height=400,
            pos=[340, 40],
            no_close=True, 
            no_collapse=True, 
            no_background=True, 
            no_resize=True, 
            no_move=True, 
            no_bring_to_front_on_focus=True, 
            no_focus_on_appearing=True,
            tag=self.window_tag,
        ):
            self.running_status_uuid = dpg.add_text("Running Status: ", color=(255, 255, 255))
            self.current_task_uuid = dpg.add_text("Current Task: ", color=(255, 255, 255))
            add_header("Queue:", color=(255, 255, 255))
            self.task_queue_uuid = dpg.generate_uuid()
            
            with dpg.group(tag=self.task_queue_uuid):
                dpg.add_text("-", color=(255, 255, 255))

            
    def update_running_status(self, status: str) -> None:
        """
        Update the running status in the GUI.
        
        Args:
            status (str): The running status of the task manager.
        """
        dpg.set_value(self.running_status_uuid, f"Running Status: {status}")
            
            
    def update_current_task(self, task_name: str) -> None:
        """
        Update the current task in the GUI.
        
        Args:
            task_name (str): The name of the current task.
        """
        dpg.set_value(self.current_task_uuid, f"Current Task: {task_name}")
        
        
    def update_task_queue(self, tasks: list) -> None:
        """
        Update the task queue in the GUI.
        
        Args:
            tasks (list): The list of tasks in the queue.
        """
        
        viewport_width = dpg.get_viewport_client_width()
        viewport_height = dpg.get_viewport_client_height()
        
        dpg.configure_item(self.window_tag, height=viewport_height-100)
        
        dpg.delete_item(self.task_queue_uuid)
        self.task_queue_uuid = dpg.generate_uuid()
        
        with dpg.group(tag=self.task_queue_uuid, parent=self.window_tag):
            for i, task in enumerate(tasks):
                with dpg.group(horizontal=True, parent=self.task_queue_uuid):
                    index_string = f'[{i}]'
                    dpg.add_text(f"{index_string:<5}", color=(255, 255, 255))
                    dpg.add_text(f"{task['name']}", color=EMPHASIS_COLOR)
                with dpg.group(horizontal=True, parent=self.task_queue_uuid):
                    dpg.add_text(f"{task['description']}", color=TEXT_COLOR, wrap=350, indent=42)
                if task['parameters']:
                    
                    with dpg.group(horizontal=True, parent=self.task_queue_uuid, indent=42):
                        
                        left_group = dpg.add_group()
                        right_group = dpg.add_group()
                    
                        for j, (param, value) in enumerate(task['parameters'].items()):
                        
                            if j%2 == 0:
                                with dpg.group(horizontal=True, parent=left_group):
                                        dpg.add_text(f"{param:<10}", color=SECONDARY_COLOR)
                                        dpg.add_text(f"{value:<10}", color=WHITE)
                            else:
                                with dpg.group(horizontal=True, parent=right_group):
                                        dpg.add_text(f"{param:<10}", color=SECONDARY_COLOR)
                                        dpg.add_text(f"{value:<10}", color=WHITE)
                            
                            

                
                dpg.add_spacer(height=5)


from loguru import logger as loguru_logger
from threading import Lock
import sys
import os
import time
from pathlib import Path
from .broadcaster import Broadcaster


class Logger(Broadcaster):
    """
    Singleton class for configuring and managing logging in the application.
    """
    
    LOG_LEVELS = {
        "DEBUG": 10,
        "INFO": 20,
        "WARNING": 30,
        "ERROR": 40,
        "EXCEPTION": 50,
    }
    
    _instance = None
    _lock = Lock()  # To make it thread-safe


    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            with cls._lock:
                if not cls._instance:
                    cls._instance = super(Logger, cls).__new__(cls, *args, **kwargs)
        return cls._instance


    def __init__(self):
        
        # Avoid reinitializing if the instance already exists
        if hasattr(self, "_initialized") and self._initialized:
            return
        super().__init__()
        self._initialized = True
        self._gui_level = "DEBUG"
        self.log_path: Path = Path()
        
    
    def _should_broadcast(self, level: str) -> bool:
        return self.LOG_LEVELS[level.upper()] >= self.LOG_LEVELS[self._gui_level.upper()]

        
        
    def configure(
        self,
        root_path: Path,
        console_level: str = "DEBUG",
        file_level: str = "DEBUG",
        gui_level: str = "DEBUG",
        file_name: Path = Path("debug.log"),
    ) -> None:
        """
        Configures the logger for the application.

        Args:
            root_path (str): The root path where the log file will be stored.
            console_level (str, optional): Logging level for console output. Defaults to "DEBUG".
            file_level (str, optional): Logging level for file output. Defaults to "DEBUG".
            file_name (str, optional): Name of the log file. Defaults to "debug.log".
        """
        self._gui_level = gui_level
        loguru_logger.remove()
        loguru_logger.add(
            sink=sys.stdout,
            colorize=True,
            level=console_level,
            #format="{time:YYYY-MM-DD at HH:mm:ss} | {level} | {message}",
        )
        log_file_path = root_path / file_name
        self.debug(f"Log file path: {log_file_path}")
        loguru_logger.add(
            sink=log_file_path,
            level=file_level,
            format="{time:YYYY-MM-DD at HH:mm:ss} | {level} | {message}",
        )
        self.info(f"Logging configured: console level={console_level}, file level={file_level}, file name={log_file_path}, gui level={self._gui_level}") 
        
        
    def info(self, message: str) -> None:
        """
        Logs an info message.

        Args:
            message (str): The message to log.
        """
        loguru_logger.info(message)
        if self._should_broadcast("INFO"):
            self.broadcast_sync({"time": time.time(), "message": message, "level": "info"})
        
        
    def debug(self, message: str) -> None:
        """
        Logs a debug message.

        Args:
            message (str): The message to log.
        """
        loguru_logger.debug(message)
        if self._should_broadcast("DEBUG"):
            self.broadcast_sync({"time": time.time(), "message": message, "level": "debug"})
        
    
    def warning(self, message: str) -> None:
        """
        Logs a warning message.

        Args:
            message (str): The message to log.
        """
        loguru_logger.warning(message)
        if self._should_broadcast("WARNING"):
            self.broadcast_sync({"time": time.time(), "message": message, "level": "warning"})
        
        
    def error(self, message: str) -> None:
        """
        Logs an error message.

        Args:
            message (str): The message to log.
        """
        loguru_logger.error(message)
        if self._should_broadcast("ERROR"):
            self.broadcast_sync({"time": time.time(), "message": message, "level": "error"})
        
        
    def exception(self, message: str) -> None:
        """
        Logs an exception message.

        Args:
            message (str): The message to log.
        """
        loguru_logger.exception(message)
        if self._should_broadcast("EXCEPTION"):
            self.broadcast({"time": time.time(), "message": message, "level": "exception"})  


# Singleton instance
logger = Logger()


import asyncio

class Broadcaster:
    """
    A class responsible for broadcasting messages to subscribed consumers.
    """

    def __init__(self):
        """
        Initialize the Broadcaster.
        """
        self._subscribers = []


    def subscribe(self, consumer):
        """
        Subscribe a consumer to this broadcaster.

        Args:
            consumer (Consumer): The consumer to subscribe.
        """
        self._subscribers.append(consumer)
        
        
    def unsubscribe(self, consumer):
        """
        Unsubscribe a consumer from this broadcaster.

        Args:
            consumer (Consumer): The consumer to unsubscribe.
        """
        self._subscribers.remove(consumer)
        

    async def broadcast(self, message):
        """
        Broadcast a message to all subscribed consumers.

        Args:
            message (Any): The message to broadcast.
        """
        for subscriber in self._subscribers:
            await subscriber.queue.put(message)
            
            
    def broadcast_sync(self, message):
        """
        Broadcast a message to all subscribed consumers synchronously.

        Args:
            message (Any): The message to broadcast.
        """
        for subscriber in self._subscribers:
            subscriber.queue.put_nowait(message)


import dearpygui.dearpygui as dpg
import asyncio
import json
from .inputs.integer_input import IntegerInput
from .inputs.string_input import StringInput
from .inputs.boolean_input import BooleanInput
from .inputs.float_input import FloatInput
from .inputs.enum_input import EnumInput
from .inputs.base_input import BaseInput
from .text import add_text, add_header, add_text_area
from .input_group import InputGroup
from ..api_client import APIClient


class EndpointPopup:
    """Class to create a popup for FastAPI endpoints."""
    
    def __init__(self, api_client, path) -> None:
        self.uuid = dpg.generate_uuid()
        self.api_client = api_client
        self.path = path
        self.input_group = InputGroup()
        self.inputs = []
        self.response_text_area = None 
        
        for param in self.path.get.parameters.values():
            input_component = self.param_to_input(param)
            self.add_input(input_component)
            
            self.input_group.add_input(input_component)
            
            
    def param_to_input(self, param: str) -> BaseInput:
        """Convert a parameter type to the corresponding input component."""
        if param.type_ == "integer":
            return IntegerInput(param.name, default_value=0)
        elif param.type_ == "string":
            return StringInput(param.name, default_value="")
        elif param.type_ == "number":
            return FloatInput(param.name, default_value=0.0)
        elif param.type_ == "boolean":
            return BooleanInput(param.name, default_value=False)
        elif param.type_ == "enum":
            return EnumInput(param.name, options=param.enum_values, default_value=param.enum_values[0])
        else:
            raise ValueError(f"Unsupported parameter type: {param.type_}")


    def add_input(self, input_component: BaseInput) -> None:
        """Add an input component to the popup."""
        self.inputs.append(input_component)


    def draw(self) -> None:
        """Draw the popup and its inputs."""
        with dpg.window(
            label=f"{self.path.get.summary}",
            width=300,
            height=-1,
            modal=False,
            show=True,
            tag=self.uuid,
        ):
            add_text(self.path.get.description, wrap=265)

            if self.input_group.length > 0:
                add_header("Parameters")
                self.input_group.draw()
                
            
            dpg.add_button(
                label="Send Request",
                callback=self.request,
                indent=10,
            )
            
            dpg.add_spacer(height=5)
            add_header("Response")
            self.response_text_area = add_text_area(height=200, width=265)
            


    def run_async(self, func) -> None:
        """Run an async function in the event loop."""
        asyncio.create_task(func)


    def get_input_data(self) -> dict:
        """Retrieve data from all inputs."""
        return {input_component.label: input_component.get_value() for input_component in self.inputs}
    
    
    def request(self) -> None:
        """Send a request to the endpoint with the input data."""
        data = self.get_input_data()
        response = self.api_client.get(endpoint=self.path.path, params=data)
        dpg.set_value(self.response_text_area, json.dumps(response, indent=2))


from .logging import logger
from .task import Task
import asyncio
    

class TaskManager:
    """
    TaskManager is a class that manages a queue of tasks. It is used by the
    Experiment class to manage the user-defined tasks."
    """

    def __init__(self):

        self._current_task: Task = None
        self._task_queue = asyncio.Queue()
        self._pause_event = asyncio.Event()
        self._pause_event.set()
        
    
    def setup(self):
        """
        Setup the task manager.
        """
        logger.debug("[TaskManager] Setup started")
        logger.debug("[TaskManager] Setup completed")
        
        
    async def run(self):
        """
        The main loop that runs the tasks in the queue.
        """
        while True:
            try:
                await self._pause_event.wait()
                logger.info("[TaskManager] Waiting for task to appear on queue")
                self._current_task = await self._task_queue.get()
                logger.info(f"[TaskManager] Task fetched from queue: {self._current_task.name}")
                try:
                    await self._current_task.start(experiment=self)
                except Exception as e:
                    logger.error(f"Error running task {self._current_task}: {e}")
                finally:
                    self._current_task = None

            except Exception as e:
                logger.error(f"Error running task manager: {e}")


    def teardown(self):
        """
        Teardown the task manager.
        """
        logger.debug("[TaskManager] Teardown started")
        logger.debug("[TaskManager] Teardown completed")
        
        
    def pause(self):
        """
        Pause the task manager.
        """
        self._pause_event.clear()
        if self._current_task:
            self._current_task.pause()
        logger.info("[TaskManager] paused.")
        
        
    def resume(self):
        """
        Resume the task manager.
        """
        self._pause_event.set()
        if self._current_task:
            self._current_task.resume()
        logger.info("[TaskManager] Resumed.")
        
        
    def abort(self):
        """
        Abort the current task.
        """
        if self._current_task:
            self._current_task.abort()
            logger.info("[TaskManager] Task manager aborted current task.")
            self.pause()
        else:
            logger.info("[TaskManager] No task to abort.")
            
            
    def current_task(self) -> Task:
        """
        Get the current task.
        """
        return self._current_task
        
        
    def add_task(self, task: Task):
        """
        Add a task to the queue.
        """
        logger.info(f"[TaskManager] Adding task to queue: {task.name}")
        self._task_queue.put_nowait(task)
        
        
    def register_task(self, task: Task):
        try:
            task.register_endpoints(self)
        except Exception as e:
            logger.error(f"Error registering endpoints for {task.__class__.__name__}: {e}")
        
        
        
        
    def register_endpoints(self, api_server):
        """
        Register the task manager endpoints with the API server.
        """
        
        @api_server.app.get("/task_manager/pause", tags=["Task Manager"])
        async def pause():
            """
            Endpoint to pause the task manager.
            """
            if not self._pause_event.is_set():
                return {"status": "success", "message": "Task manager is already paused."}      
            self.pause()
            return {"status": "success", "message": "Task manager paused."}
        
        
        @api_server.app.get("/task_manager/resume", tags=["Task Manager"])
        async def resume():
            """
            Endpoint to resume the task manager.
            """
            if self._pause_event.is_set():
                return {"status": "success", "message": "Task manager is already running."}    
            self.resume()
            return {"status": "success", "message": "Task manager resumed."}
        
        
        @api_server.app.get("/task_manager/abort", tags=["Task Manager"])
        async def abort():
            """
            Endpoint to abort the current task.
            """
            if not self._current_task:
                return {"status": "success", "message": "No task to abort."}
            self.abort()
            return {"status": "success", "message": "Task manager aborted current task."}
        
        
        @api_server.app.get("/task_manager/status", tags=["Task Manager"])
        async def status():
            """
            Endpoint to get the status of the task manager.
            """
            if self._pause_event.is_set():
                return {
                    "status": 200,
                    "data": "Running",
                }
            else:
                return {
                    "status": 200,
                    "data": "Paused",
                }
        
        
        @api_server.app.get("/task_manager/current_task", tags=["Task Manager"])
        async def current_task():
            """
            Endpoint to get the current task.
            """
            if self._current_task:
                return {
                    "status": 200,
                    "data": f"{self._current_task.name}",
                }
            else:
                return {
                    "status": 200,
                    "data": None,
                }
        
        
        @api_server.app.get("/task_manager/add_task", tags=["Task Manager"])
        async def add_task():
            """
            Endpoint to add a task to the queue.
            """
            from ..tasks import TestTask
            logger.info(f"Adding task to queue")
            self.add_task(TestTask())
            
        
        @api_server.app.get("/task_manager/task_list", tags=["Task Manager"])
        async def task_list():
            """
            Endpoint to get the list of tasks in the queue.
            """
            tasks = []
            for task in self._task_queue._queue:
                tasks.append(task.display_dict())
            return {
                "status": 200,
                "data": tasks,
            }


import dearpygui.dearpygui as dpg
from ...core.consumer import Consumer
from ...core.logging import logger
import time
import asyncio



class LivePlotWidget(Consumer):
    
    
    def __init__(self, data: dict[str, list]):
        super().__init__()
        
        self.window_tag = dpg.generate_uuid()
        self.plot_tag = dpg.generate_uuid()
        self.x_axis_tag = dpg.generate_uuid()
        self.y_axis_tag = dpg.generate_uuid()
        self.data = data.copy()
        self.x_key = next(iter(self.data.keys()))
        self.series_tags = {}
        
        with dpg.window(
            label="Live Plot",
            pos=[20, 40],
            width=800,
            height=600,
            tag=self.window_tag,
        ):
            with dpg.menu_bar():
                dpg.add_menu_item(label="Clear", callback=self.clear_data)
                with dpg.menu(label="x-axis"):
                    for key in self.data.keys():
                        dpg.add_menu_item(label=key, callback=lambda s, a, u: self.set_x_key(s, a, u), user_data=key)
                
            
            with dpg.plot(label="Live Data Plot", tag=self.plot_tag, height=-1, width=-1):
                dpg.add_plot_legend()
                dpg.add_plot_axis(dpg.mvXAxis, label=self.x_key, tag=self.x_axis_tag)
                dpg.add_plot_axis(dpg.mvYAxis, label="Value", tag=self.y_axis_tag)

        self.build_series()
        
        
    def clear_data(self):
        """
        Clear the data in the plot.
        """
        for key in self.data.keys():
            self.data[key] = []
        
        
    def set_on_close(self, callback: callable):
        """
        Set the callback for when the window is closed.
        """
        dpg.configure_item(self.window_tag, on_close=callback)
        
        
    def set_x_key(self, sender, app_data, user_data):
        """
        Set the x-axis key for the plot.
        """
        if user_data in self.data:
            self.x_key = user_data
            dpg.configure_item(self.x_axis_tag, label=user_data)
        else:
            logger.error(f"Key {user_data} not found in data.")
        
    
    def build_series(self):
        """
        Build the series for the plot.
        """
        for key in self.data.keys():
            series_tag = dpg.generate_uuid()
            dpg.add_scatter_series(
                x=self.data[self.x_key],
                y=self.data[key],
                label=key,
                tag=series_tag,
                parent=self.y_axis_tag,
            )
            self.series_tags[key] = series_tag

    
    def update_data(self, data: dict):
        """
        Update the plot with new data.
        """
        for key, value in data.items():
            if key in self.data:
                self.data[key].append(value)
                
                
    def update_series(self):
        """
        Update the series in the plot.
        """
        for key in self.data:
            dpg.set_value(self.series_tags[key], [self.data[self.x_key], self.data[key]])
            
            
    def update(self, data: dict):
        """
        Update the plot with new data.
        """
        self.update_data(data)
        self.update_series()
    